from django import forms
from django.contrib.auth import get_user_model
from django.db.models import Q
from django.forms import widgets
from django.utils.translation import ugettext_lazy as _

from viralhostrangedb import models, data_file_reader


class ImportDataSourceForm(forms.ModelForm):
    file = forms.FileField(
        label=_('File'),
        help_text=_("The file to upload."),
        required=False,
    )
    url = forms.URLField(
        label=_('URL'),
        help_text=_("A link from which the data source can be downloaded."),
        required=False,
    )

    class Meta:
        model = models.DataSource
        fields = (
            'name',
            'description',
            'public',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.on_disk_file = None

    def full_clean(self):
        super().full_clean()
        if not self.is_bound:  # Stop further processing.
            return
        if (self.files is None or len(self.files) == 0) and len(self.cleaned_data["url"]) == 0 or \
                (self.files is not None and len(self.files) > 0) and len(self.cleaned_data["url"]) > 0:
            self.add_error("url", _("You have to either provide a file or an URL, and not both."))
            self.add_error("file", _("You have to either provide a file or an URL, and not both."))

    def save(self, owner=None, commit=True):
        instance = super().save(commit=False)
        instance.owner = owner
        instance.save()
        file = None
        if self.files is not None and len(self.files) == 1:
            file = self.files['file']
            instance.raw_name = file.name
            instance.kind = "FILE"
        if len(self.cleaned_data["url"]) > 0:
            try:
                instance.kind = "URL"
                url = self.cleaned_data["url"]
                instance.raw_name = url
                raise NotImplementedError("Downloading from an URL is not available yet")
            except KeyError:
                pass
        if file is not None:
            # with NamedTemporaryFile(mode='wb+', suffix="-%s" % file.name, delete=False) as destination:
            #     print(destination.name)
            #     for chunk in file.chunks():
            #         destination.write(chunk)
            #     file = destination.name

            # with open(file, "rb") as input_file:
            data_file_reader.import_file(
                file=file,
                data_source=instance,
            )
        if commit:
            instance.save()
        return instance


class DataSourceUserUpdateForm(forms.ModelForm):
    new_allowed_users = forms.CharField(
        label=_("new_allowed_users__label"),
        help_text=_("new_allowed_users__help_text"),
        widget=widgets.Textarea(),
        required=False,
    )

    class Meta:
        model = models.DataSource
        fields = (
            'name',
            'description',
            'public',
            'allowed_users',
        )
        widgets = dict(
            raw_name=widgets.TextInput(),
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["allowed_users"].queryset = self.instance.allowed_users

    def save(self, commit=True):
        instance = super().save(commit=commit)
        candidates = [c.strip() for c in
                      self.cleaned_data["new_allowed_users"].replace("\r", "\n").replace("\n\n", "\n").split("\n")]

        try:
            # important to prevent allowing each user with empty email at each save
            candidates.remove('')
        except ValueError:
            pass

        for candidate in get_user_model().objects.filter(Q(username__in=candidates) | Q(email__in=candidates)).only(
                'pk'):
            instance.allowed_users.add(candidate)
        if commit:
            instance.save()
        return instance


class MappingForm(forms.Form):
    raw_response = forms.FloatField(
    )
    mapping = forms.ModelChoiceField(
        queryset=models.GlobalViralStrainResponseValue.objects_mappable(),
        required=True,
    )


MappingFormSet = forms.formset_factory(
    form=MappingForm,
    extra=0,
    can_delete=False,
)
