from django.contrib.auth.mixins import AccessMixin
from django.db.models import Q
from django.http import HttpResponseForbidden, Http404


class OnlyPublicOrOwnedMixin(AccessMixin):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return super().get_queryset().filter(public=True)
        return super().get_queryset().filter(Q(owner=self.request.user) | Q(public=True))

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request=request, *args, **kwargs)
        except Http404 as e:
            if self.request.user.is_authenticated:
                raise e
            return self.handle_no_permission()


class OnlyPublicOrOwnedOrStaffMixin(AccessMixin):
    def get_queryset(self):
        if self.request.user.is_staff:
            return super().get_queryset()
        if not self.request.user.is_authenticated:
            return super().get_queryset().filter(public=True)
        return super().get_queryset().filter(Q(owner=self.request.user) | Q(public=True))

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request=request, *args, **kwargs)
        except Http404 as e:
            if self.request.user.is_authenticated:
                raise e
            return self.handle_no_permission()


class StaffOnlyMixin(AccessMixin):
    def dispatch(self, request, *args, **kwargs):
        if not self.request.user.is_authenticated:
            return self.handle_no_permission()
        if not request.user.is_staff:
            return HttpResponseForbidden()
        return super().dispatch(request=request, *args, **kwargs)


class OnlyOwnedMixin(AccessMixin):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return super().get_queryset().none()
        return super().get_queryset().filter(owner=self.request.user)

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request=request, *args, **kwargs)
        except Http404 as e:
            if self.request.user.is_authenticated:
                raise e
            return self.handle_no_permission()


class OnlyOwnedOrStaffMixin(AccessMixin):
    def get_queryset(self):
        if self.request.user.is_staff:
            return super().get_queryset()
        if not self.request.user.is_authenticated:
            return super().get_queryset().none()
        return super().get_queryset().filter(owner=self.request.user)

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request=request, *args, **kwargs)
        except Http404 as e:
            if self.request.user.is_authenticated:
                raise e
            return self.handle_no_permission()
