from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.validators import MaxLengthValidator
from django.db import models
# Create your models here.
from django.db.models import Exists, OuterRef, Q
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class GlobalViralStrainResponseValue(models.Model):
    class Meta:
        verbose_name = _("Global Viral-Strain response values")
        verbose_name_plural = _("Global Viral-Strain response values")

    name = models.CharField(
        verbose_name=_("response__name__verbose_name"),
        help_text=_("response__name__help_text"),
        max_length=32,
        validators=[MaxLengthValidator(32)] if settings.DEBUG else [],
        unique=True,
    )
    description = models.TextField(
        verbose_name=_("response__description__verbose_name"),
        help_text=_("response__description__help_text"),
        blank=True,
        default="",
    )
    value = models.FloatField(
        verbose_name=_("response__value__verbose_name"),
        help_text=_("response__value__help_text"),
        default=-1000000,
    )

    def __str__(self):
        return "%s (value=%s)" % (self.name, '{0:g}'.format(self.value))

    @staticmethod
    def get_not_mapped_yet_keyword():
        return 'NOT MAPPED YET'

    @staticmethod
    def get_not_mapped_yet():
        return GlobalViralStrainResponseValue.objects.get(
            name=GlobalViralStrainResponseValue.get_not_mapped_yet_keyword())

    @classmethod
    def objects_mappable(cls):
        return cls.objects.filter(~Q(name=cls.get_not_mapped_yet_keyword()))


class DataSourceWithPendingManager(models.Manager):
    def get_queryset(self):
        pending = super().get_queryset().filter(
            responseindatasource__response=GlobalViralStrainResponseValue.get_not_mapped_yet()
        ).filter(pk=OuterRef('pk'))
        return super().get_queryset().annotate(
            mapping_done=~Exists(pending)
        )


class DataSource(models.Model):
    class Meta:
        verbose_name = _("Data Source")
        verbose_name_plural = _("Data Sources")

    objects_annotated_with_pending = DataSourceWithPendingManager()

    objects = models.Manager()

    owner = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.PROTECT,
    )
    public = models.BooleanField(
        verbose_name=_("public__verbose_name"),
        help_text=_("public__help_text"),
        default=False,
    )
    name = models.CharField(
        verbose_name=_("data_source__name__verbose_name"),
        help_text=_("data_source__name__help_text"),
        max_length=128,
        validators=[MaxLengthValidator(128)] if settings.DEBUG else []
    )
    description = models.TextField(
        verbose_name=_("data_source__description__verbose_name"),
        help_text=_("data_source__description__help_text"),
        blank=True,
        default="",
    )
    raw_name = models.TextField(
        verbose_name=_("raw_name__verbose_name"),
        help_text=_("raw_name__help_text"),
    )
    kind = models.CharField(
        verbose_name=_("kind__verbose_name"),
        help_text=_("kind__help_text"),
        max_length=4,
        choices=(
            ("FILE", _("From a file")),
            ("URL", _("From an URL")),
        )
    )
    allowed_users = models.ManyToManyField(
        to=get_user_model(),
        verbose_name=_("allowed_users__verbose_name"),
        help_text=_("allowed_users__help_text"),
        related_name='datasource_readable_set',
        blank=True,
    )
    creation_date = models.DateTimeField(
        verbose_name=_("creation_date__verbose_name"),
        help_text=_("creation_date__help_text"),
    )
    last_edition_date = models.DateTimeField(
        verbose_name=_("last_edition_date__verbose_name"),
        help_text=_("last_edition_date__help_text"),
    )

    def __str__(self):
        return "%s (created on %s)" % (
            self.name,
            "???" if self.creation_date is None else self.creation_date.strftime('%Y-%m-%d %H:%M:%S %Z'),
        )

    def get_absolute_url(self):
        return reverse("viralhostrangedb:data-source-detail", args=[self.pk])

    def get_mapping(self):
        used_responses = dict()
        for r in GlobalViralStrainResponseValue.objects.filter(
                pk__in=self.responseindatasource.values_list('response__pk').distinct()):
            used_responses[r.pk] = r
        for raw_response, response_mapped in self.responseindatasource. \
                values_list('raw_response', 'response__pk').distinct():
            yield raw_response, used_responses[response_mapped]

    @property
    def is_mapping_done(self):
        return not GlobalViralStrainResponseValue.objects \
            .filter(responseindatasource__data_source__pk=self.pk) \
            .filter(name=GlobalViralStrainResponseValue.get_not_mapped_yet_keyword()) \
            .exists()

    @classmethod
    def get_not_mapped_data_sources(cls, *, owner):
        return cls.objects.filter(pk__in=cls.get_not_mapped_data_sources_pk(owner))

    @classmethod
    def get_not_mapped_data_sources_pk(cls, *, owner):
        return cls._get_not_mapped_data_sources(owner=owner).distinct()

    @classmethod
    def has_not_mapped_data_sources(cls, *, owner):
        return cls._get_not_mapped_data_sources(owner=owner).exists()

    @classmethod
    def _get_not_mapped_data_sources(cls, *, owner=None):
        q = GlobalViralStrainResponseValue.get_not_mapped_yet() \
            .responseindatasource \
            .values("data_source__pk")
        if owner:
            q = q.filter(data_source__owner__pk=owner.pk)
        return q

    def save(self, *args, **kwargs):
        if self.pk is None:
            self.creation_date = timezone.now()
        self.last_edition_date = timezone.now()
        super().save(*args, **kwargs)


class Strain(models.Model):
    class Meta:
        verbose_name = _("Strain")
        verbose_name_plural = _("Strains")

    data_source = models.ManyToManyField(
        to=DataSource,
        verbose_name=_("data_source__verbose_name"),
        help_text=_("data_source__help_text"),
    )
    name = models.CharField(
        verbose_name=_("strain__name__verbose_name"),
        help_text=_("strain__name__help_text"),
        max_length=64,
        validators=[MaxLengthValidator(64)] if settings.DEBUG else []
    )

    def __str__(self):
        return "%s from %s" % (self.name, ", ".join(str(s) for s in self.data_source.all()))


class Virus(models.Model):
    class Meta:
        verbose_name = _("Virus")
        verbose_name_plural = _("Viruses")

    data_source = models.ManyToManyField(
        to=DataSource,
        verbose_name=_("data_source__verbose_name"),
        help_text=_("data_source__help_text"),
    )
    name = models.CharField(
        verbose_name=_("virus__name__verbose_name"),
        help_text=_("virus__name__help_text"),
        max_length=64,
        validators=[MaxLengthValidator(64)] if settings.DEBUG else []
    )

    def __str__(self):
        return "%s from %s" % (self.name, ", ".join(str(s) for s in self.data_source.all()))


class ViralStrainResponseValueInDataSource(models.Model):
    class Meta:
        verbose_name = _("Viral-Strain response value in data sources")
        verbose_name_plural = _("Viral-Strain responses value in data sources")

    data_source = models.ForeignKey(
        to=DataSource,
        verbose_name=_("data_source__verbose_name"),
        help_text=_("data_source__help_text"),
        on_delete=models.CASCADE,
        related_name="responseindatasource"
    )
    virus = models.ForeignKey(
        to=Virus,
        verbose_name=_("virus__verbose_name"),
        help_text=_("virus__help_text"),
        on_delete=models.CASCADE,
        related_name="responseindatasource"
    )
    strain = models.ForeignKey(
        to=Strain,
        verbose_name=_("strain__verbose_name"),
        help_text=_("strain__help_text"),
        on_delete=models.CASCADE,
        related_name="responseindatasource"
    )
    raw_response = models.FloatField(
        verbose_name=_("raw_response__verbose_name"),
        help_text=_("raw_response__help_text"),
    )
    response = models.ForeignKey(
        to=GlobalViralStrainResponseValue,
        verbose_name=_("response__verbose_name"),
        help_text=_("response__help_text"),
        on_delete=models.PROTECT,
        related_name="responseindatasource"
    )

    def __str__(self):
        return "(%s,%s)%s" % (self.virus.name, self.strain.name, self.raw_response)
