import json
import logging
import os

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from viralhostrangedb import data_file_reader, models

logger = logging.getLogger(__name__)


class DataFileReaderTestCase(TestCase):
    test_data = "./test_data"

    def setUp(self):
        self.owner = get_user_model().objects.create(
            username="root",
        )

    def test_read_work_fine(self):
        for f in [f for f in os.listdir(self.test_data) if f.endswith(".xlsx") and not f.endswith(".err.xlsx")]:
            vhrs = data_file_reader.parse_file(os.path.join(self.test_data, f))
            vhrs_dict = data_file_reader.to_dict(vhrs)
            try:
                with open(os.path.join(self.test_data, f + ".json")) as f:
                    data = json.load(f)
                    self.assertEqual(data, vhrs_dict)
            except FileNotFoundError as e:
                logger.exception("json associated file is missing, reducing robustness of tests. "
                                 "We created it, check if it is correct")
                logger.exception(json.dumps(vhrs_dict, indent=2))
                with open(os.path.join(self.test_data, f + ".json.candidate"), "w") as f:
                    f.write(json.dumps(vhrs_dict, indent=2))

    def test_read_work_with_stream(self):
        with open(os.path.join(self.test_data, "simple.xlsx"), "rb") as input_file:
            data_file_reader.parse_file(input_file)

    def test_import_in_db(self):
        filename = os.path.join(self.test_data, "simple.xlsx")
        data_source, created = models.DataSource.objects.get_or_create(
            raw_name=filename,
            kind="FILE",
            owner=self.owner,
            creation_date=timezone.now(),
            last_edition_date=timezone.now(),
        )
        data_file_reader.import_file(
            data_source=data_source,
            file=filename,
        )
        self.assertEqual(data_source.owner, self.owner)
        self.assertEqual(models.Virus.objects.count(), 3)
        self.assertEqual(models.Strain.objects.count(), 4)
        self.assertEqual(models.ViralStrainResponseValueInDataSource.objects.get(
            data_source=data_source,
            strain=models.Strain.objects.get(data_source=data_source, name="C4"),
            virus=models.Virus.objects.get(data_source=data_source, name="r1"),
        ).raw_response, 4)
        self.assertEqual(models.ViralStrainResponseValueInDataSource.objects.get(
            data_source=data_source,
            strain=models.Strain.objects.get(data_source=data_source, name="C2"),
            virus=models.Virus.objects.get(data_source=data_source, name="r2"),
        ).raw_response, 6)
        self.assertEqual(models.ViralStrainResponseValueInDataSource.objects.get(
            data_source=data_source,
            strain=models.Strain.objects.get(data_source=data_source, name="C3"),
            virus=models.Virus.objects.get(data_source=data_source, name="r3"),
        ).raw_response, 11)
        self.assertEqual(models.ViralStrainResponseValueInDataSource.objects.get(
            data_source=data_source,
            strain=models.Strain.objects.get(data_source=data_source, name="C1"),
            virus=models.Virus.objects.get(data_source=data_source, name="r1"),
        ).raw_response, 1)
