import logging
import os

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models import Count
from django.test import TestCase
from django.urls import reverse

from viralhostrangedb import models, data_file_reader

logger = logging.getLogger(__name__)


class ViewTestCase(TestCase):
    test_data = "./test_data"

    def setUp(self):
        self.user_pwd = "eil2guj4cuSho2Vai3hu"
        get_user_model().objects.create(
            username="tata",
        )
        get_user_model().objects.create(
            username="titi",
        )

        self.user = get_user_model().objects.create(
            username="user",
            email="b@a.a",
            is_staff=True,
        )
        self.user.set_password(self.user_pwd)
        self.user.save()

        self.toto = get_user_model().objects.create(
            username="toto",
            email="a@a.a",
        )
        self.toto.set_password(self.user_pwd)
        self.toto.save()

        self.private_data_source_of_user = models.DataSource.objects.create(
            owner=self.user,
            public=False,
            name="private user",
            raw_name="ee",
            kind="FILE",
        )
        filename = os.path.join(self.test_data, "simple.xlsx")
        data_file_reader.import_file(
            data_source=self.private_data_source_of_user,
            file=filename,
        )
        self.public_data_source_of_user = models.DataSource.objects.create(
            owner=self.user,
            public=True,
            name="public user",
            raw_name="ee",
            kind="FILE",
        )
        self.public_data_source_of_user.allowed_users.add(self.user)
        self.public_data_source_of_user.save()
        filename = os.path.join(self.test_data, "three_reponse_simple.xlsx")
        data_file_reader.import_file(
            data_source=self.public_data_source_of_user,
            file=filename,
        )
        self.private_data_source_of_toto = models.DataSource.objects.create(
            owner=self.toto,
            public=False,
            name="private toto",
            raw_name="ee",
            kind="FILE",
        )
        filename = os.path.join(self.test_data, "simple.xlsx")
        data_file_reader.import_file(
            data_source=self.private_data_source_of_toto,
            file=filename,
        )


class HomePageTestCase(ViewTestCase):
    def test_works(self):
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(reverse('viralhostrangedb:home'))
        self.assertEqual(response.status_code, 200)


class FileImportTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:file-import-view')
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_import_work(self):
        url = reverse('viralhostrangedb:file-import-view')
        self.client.force_login(self.toto)

        f = open(os.path.join(self.test_data, "three_reponse_simple.xlsx"), "rb")

        form_data = dict(
            name='imported',
            description="Nop",
            public=True,
            file=SimpleUploadedFile(f.name, f.read(), content_type="application/vnd.ms-excel"),
        )

        response = self.client.post(url, form_data, follow=True)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(self.toto.datasource_set.filter(name="imported").count(), 1)
        self.assertEqual(self.toto.datasource_set.get(name="imported").responseindatasource.count(), 12)


class DataSourceMappingPendingListViewTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:data-source-pending-mapping-list')
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)


class BrowseTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:browse')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)


class DataSourceListViewTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:data-source-list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)


class DataSourceDeleteViewTestCase(ViewTestCase):

    def test_works(self):
        url = reverse('viralhostrangedb:data-source-delete', args=[self.private_data_source_of_user.pk])
        strain_pk = list(self.private_data_source_of_user.strain_set.values_list('pk', flat=True))
        virus_pk = list(self.private_data_source_of_user.virus_set.values_list('pk', flat=True))
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.client.force_login(self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response = self.client.post(url, {})
        self.assertEqual(response.status_code, 302)
        # check that the object is not here anymore
        self.assertFalse(models.DataSource.objects.filter(pk=self.private_data_source_of_user.pk).exists())
        # check that virus only used by it are removed
        self.assertEqual(
            models.Virus.objects
                .filter(pk__in=virus_pk)
                .annotate(cpt=Count('data_source'))
                .filter(cpt=0)
                .count(),
            0,
        )
        # check that strain only used by it are removed
        self.assertEqual(
            models.Strain.objects
                .filter(pk__in=strain_pk)
                .annotate(cpt=Count('data_source'))
                .filter(cpt=0)
                .count(),
            0,
        )

    def test_cross_user_block(self):
        url = reverse('viralhostrangedb:data-source-delete', args=[self.private_data_source_of_user.pk])
        self.client.force_login(self.toto)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)


class DataSourceUpdateViewTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.private_data_source_of_user.pk])
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_cross_users(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.private_data_source_of_user.pk])
        self.client.force_login(self.toto)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_staff_can_not_edit_private(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.private_data_source_of_toto.pk])
        self.client.force_login(self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_mapping_work(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.public_data_source_of_user.pk])
        self.client.force_login(self.user)

        former_name = self.public_data_source_of_user.name
        allowed_users = set([u.pk for u in self.public_data_source_of_user.allowed_users.all()])

        form_data = dict(
            name=former_name + former_name,
            description="",
            new_allowed_users="",
            allowed_users=",".join([str(u.pk) for u in self.public_data_source_of_user.allowed_users.all()]),
        )

        response = self.client.post(url, form_data)
        self.assertRedirects(
            response,
            expected_url=reverse('viralhostrangedb:data-source-detail', args=[self.public_data_source_of_user.pk])
        )

        self.assertEqual(
            models.DataSource.objects.get(pk=self.public_data_source_of_user.pk).name,
            former_name + former_name
        )
        self.assertSetEqual(allowed_users, set([u.pk for u in self.public_data_source_of_user.allowed_users.all()]))

    def test_new_user_granted(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.public_data_source_of_user.pk])
        self.client.force_login(self.user)

        allowed_users = set([u.pk for u in self.public_data_source_of_user.allowed_users.all()])

        form_data = dict(
            name=self.public_data_source_of_user.name,
            description="",
            new_allowed_users="toto",
            allowed_users=",".join([str(u.pk) for u in self.public_data_source_of_user.allowed_users.all()]),
        )
        allowed_users.add(self.toto.pk)

        self.client.post(url, form_data)
        self.assertSetEqual(allowed_users, set([u.pk for u in self.public_data_source_of_user.allowed_users.all()]))

    def test_user_injection_prevented(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.public_data_source_of_user.pk])
        self.client.force_login(self.user)

        allowed_users = set([u.pk for u in self.public_data_source_of_user.allowed_users.all()])

        form_data = dict(
            name=self.public_data_source_of_user.name,
            description="",
            new_allowed_users="",
            allowed_users=",".join([str(u.pk) for u in get_user_model().objects.all()]),
        )

        self.client.post(url, form_data)
        self.assertSetEqual(allowed_users, set([u.pk for u in self.public_data_source_of_user.allowed_users.all()]))

    def test_cross_user_edition_prevented(self):
        url = reverse('viralhostrangedb:data-source-update', args=[self.private_data_source_of_toto.pk])
        self.client.force_login(self.user)

        former_description = self.private_data_source_of_toto.description
        form_data = dict(
            name=self.private_data_source_of_toto.name,
            description="XXX",
            new_allowed_users="",
            allowed_users=",".join([str(u.pk) for u in get_user_model().objects.all()]),
        )

        response = self.client.post(url, form_data)
        self.assertNotEqual(response.status_code, 200)

        self.assertEqual(
            models.DataSource.objects.get(pk=self.private_data_source_of_toto.pk).description,
            former_description
        )


class DataSourceDetailViewTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:data-source-detail', args=[self.private_data_source_of_user.pk])
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.assertTrue(self.client.login(username="user", email="a@a.a", password=self.user_pwd))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_cross_users(self):
        url = reverse('viralhostrangedb:data-source-detail', args=[self.private_data_source_of_user.pk])
        self.client.force_login(self.toto)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_staff_can_see_private(self):
        url = reverse('viralhostrangedb:data-source-detail', args=[self.private_data_source_of_toto.pk])
        self.client.force_login(self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_public_can_be_seen(self):
        url = reverse('viralhostrangedb:data-source-detail', args=[self.public_data_source_of_user.pk])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.client.force_login(self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.client.force_login(self.toto)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)


class BrowseDataSetTestCase(ViewTestCase):
    def test_works(self):
        response = self.client.get(
            reverse('viralhostrangedb:data-source-browse', args=[self.public_data_source_of_user.pk]))
        viruses = ",".join(str(i) for i in self.public_data_source_of_user.virus_set.values_list('pk', flat=True))
        strains = ",".join(str(i) for i in self.public_data_source_of_user.strain_set.values_list('pk', flat=True))

        url = "%(url)s?strains=%(strains)s&viruses=%(viruses)s" % dict(
            url=reverse("viralhostrangedb:browse"),
            viruses=viruses,
            strains=strains,
        )
        self.assertRedirects(
            response,
            expected_url=url
        )


class DataSourceMappingEditTestCase(ViewTestCase):
    def test_works(self):
        url = reverse('viralhostrangedb:data-source-mapping-edit', args=[self.private_data_source_of_toto.pk])
        response = self.client.get(url)
        self.assertRedirects(
            response,
            expected_url=reverse('basetheme_bootstrap:login') + "?next=" + url
        )
        self.client.force_login(self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.client.force_login(self.toto)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_mapping_work(self):
        url = reverse('viralhostrangedb:data-source-mapping-edit', args=[self.public_data_source_of_user.pk])
        self.client.force_login(self.user)

        self.assertFalse(self.public_data_source_of_user.is_mapping_done)

        form_data = {
            "form-TOTAL_FORMS": 3,
            "form-INITIAL_FORMS": 3,
            "form-MIN_NUM_FORMS": 0,
            "form-MAX_NUM_FORMS": 1000,
            "form-0-raw_response": 0.0,
            "form-0-mapping": 3,
            "form-1-raw_response": 1.0,
            "form-1-mapping": 3,
            "form-2-raw_response": 2.0,
            "form-2-mapping": 2
        }

        response = self.client.post(url, form_data)
        self.assertRedirects(
            response,
            expected_url=reverse('viralhostrangedb:data-source-detail', args=[self.public_data_source_of_user.pk])
        )

        self.assertTrue(self.public_data_source_of_user.is_mapping_done)
        raw_to_response = {}
        for raw_response, response in self.public_data_source_of_user.get_mapping():
            raw_to_response[raw_response] = response
        self.assertEqual(raw_to_response.pop(0.0), models.GlobalViralStrainResponseValue.objects.get(name="No lysis"))
        self.assertEqual(raw_to_response.pop(1.0), models.GlobalViralStrainResponseValue.objects.get(name="No lysis"))
        self.assertEqual(raw_to_response.pop(2.0), models.GlobalViralStrainResponseValue.objects.get(name="Lysis"))
        self.assertEqual(len(raw_to_response), 0)

    def test_mapping_form_incomplet(self):
        url = reverse('viralhostrangedb:data-source-mapping-edit', args=[self.public_data_source_of_user.pk])
        self.client.force_login(self.user)

        self.assertFalse(self.public_data_source_of_user.is_mapping_done)

        form_data = {}

        self.assertRaises(ValidationError, self.client.post, url, form_data)

        self.assertFalse(self.public_data_source_of_user.is_mapping_done)
