from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db import transaction
from django.db.models import Count
from django.http import HttpResponseRedirect, HttpResponseForbidden
from django.shortcuts import render, get_object_or_404, redirect
from django.urls import reverse, reverse_lazy
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django.views.generic import ListView, DetailView, UpdateView, DeleteView

from viralhostrangedb import forms
from viralhostrangedb import models
from viralhostrangedb.mixins import OnlyPublicOrOwnedMixin, OnlyOwnedMixin, OnlyPublicOrOwnedOrStaffMixin


def index(request):
    context = dict(
    )
    return render(
        request=request,
        template_name='viralhostrangedb/index.html',
        context=context,
    )


@login_required
def file_import(request):
    if request.method == 'POST':
        form = forms.ImportDataSourceForm(request.POST, request.FILES)
        if form.is_valid():
            data_source = form.save(owner=request.user)
            url = reverse("viralhostrangedb:data-source-mapping-edit", args=[data_source.pk])
            messages.success(
                request=request,
                message=mark_safe(_(
                    "Successfully imported data source <b>%(data_source_name)s</b>."
                    """Redirecting to <a href="%(url)s">%(url)s</a> to map the responses to the global scheme."""
                ) % dict(
                    data_source_name=data_source.name,
                    url=url
                )),
            )
            return HttpResponseRedirect(url)
    else:
        form = forms.ImportDataSourceForm()
    context = dict(
        title=_("Import a file in the databases"),
        form=form,
    )
    return render(
        request=request,
        template_name='basetheme_bootstrap/small_form_host.html',
        context=context,
    )


@login_required
def data_source_mapping_edit(request, pk):
    data_source = get_object_or_404(models.DataSource, pk=pk)
    if request.user != data_source.owner:
        return HttpResponseForbidden()
    if request.method == 'POST':
        form = forms.MappingFormSet(request.POST, request.FILES)
        if form.is_valid():
            with transaction.atomic():
                for mapping_choice in form.forms:
                    models.ViralStrainResponseValueInDataSource.objects.filter(
                        data_source=data_source,
                        raw_response=mapping_choice.cleaned_data["raw_response"]
                    ).update(response=mapping_choice.cleaned_data["mapping"])
            return HttpResponseRedirect(reverse("viralhostrangedb:data-source-detail", args=[data_source.pk]))

    else:
        initial = []
        for raw_response, response_mapped in data_source.responseindatasource. \
                values_list('raw_response', 'response__pk').distinct():
            initial.append(
                dict(
                    raw_response=raw_response,
                    mapping=response_mapped
                )
            )
        form = forms.MappingFormSet(initial=initial)
    context = dict(
        title=_("Import a file in the databases"),
        form=form,
        medium_width=True,
        formset_css_classes="mb-4 col-12 col-md-6 col-lg-4 col-xl-3",
    )
    return render(
        request=request,
        template_name='basetheme_bootstrap/form_host.html',
        context=context,
    )


class DataSourceDetailView(OnlyPublicOrOwnedOrStaffMixin, DetailView):
    model = models.DataSource


class DataSourceUpdateView(OnlyOwnedMixin, UpdateView):
    model = models.DataSource
    form_class = forms.DataSourceUserUpdateForm
    template_name = 'basetheme_bootstrap/small_form_host.html'

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["page_title"] = _("Editing a data source")
        context["title"] = _("Editing a data source")
        return context


class DataSourceDeleteView(OnlyOwnedMixin, DeleteView):
    model = models.DataSource
    form_class = forms.DataSourceUserUpdateForm
    template_name = 'basetheme_bootstrap/small_form_host.html'
    success_url = reverse_lazy('viralhostrangedb:home')

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        page_title = mark_safe('%s "%s" <small><i>created on %s</i></small>' % (
            _("Data source"),
            self.object.name,
            self.object.creation_date.strftime('%Y-%m-%d %H:%M:%S %Z')
        ))
        context.update(dict(
            page_title=page_title,
            title=_("Deleting a data source"),
            submit_text=_('Delete data source and all depending data'),
            medium_width=True,
            btn_classes='btn-lg btn-danger text-center',
            btn_container_classes='text-center',
        ))
        return context

    @transaction.atomic
    def delete(self, request, *args, **kwargs):
        obj = self.get_object()
        strain_pk = list(obj.strain_set.values_list('pk', flat=True))
        virus_pk = list(obj.virus_set.values_list('pk', flat=True))
        models.Virus.objects \
            .filter(pk__in=virus_pk) \
            .annotate(cpt=Count('data_source')) \
            .filter(cpt=1) \
            .delete()
        models.Strain.objects \
            .filter(pk__in=strain_pk) \
            .annotate(cpt=Count('data_source')) \
            .filter(cpt=1) \
            .delete()
        return super().delete(request=request, *args, **kwargs)


class DataSourceListView(OnlyPublicOrOwnedMixin, ListView):
    model = models.DataSource
    queryset = models.DataSource.objects_annotated_with_pending

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["page_title"] = _("Data sources")
        context["title"] = _("Data sources")
        return context


class DataSourceMappingPendingListView(LoginRequiredMixin, DataSourceListView):
    def get_queryset(self):
        return super().get_queryset().filter(mapping_done=False)

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["page_title"] = _("Data sources with mapping not completed")
        context["title"] = _("Data sources with mapping not completed")
        return context


class SelectDataFromDataSourceView(OnlyPublicOrOwnedMixin, DetailView):
    model = models.DataSource

    def get(self, request, *args, **kwargs):
        data_source = self.get_object()
        viruses = ",".join(str(i) for i in data_source.virus_set.values_list('pk', flat=True))
        strains = ",".join(str(i) for i in data_source.strain_set.values_list('pk', flat=True))
        return redirect("%(url)s?strains=%(strains)s&viruses=%(viruses)s" % dict(
            url=reverse("viralhostrangedb:browse"),
            viruses=viruses,
            strains=strains,
        ), permanent=False, )
